class RootDirIsNotADirectoryError(Exception):
  ...

class RootDirDoesNotExistError(Exception):
  ...

class MetadataFileDoesNotExistError(Exception):
  ...

class InvalidMetadataFileError(Exception):
  ...